#include "distribution.h"

Distribution * ReadDistribution(istream & modelFile)
{
  string s;
  modelFile >> s;
  if (s == "OM_BINARY") {
    return new BinaryDist(modelFile);
  } else if (s == "OM_GAUSSIAN") {
    return new GaussianDist(modelFile);
  } else if (s == "OM_EXPONENTIAL") {
    return new ExponentialDist(modelFile);
  } else if (s == "OM_MIXTURE") {
    return new MixtureDist(modelFile);
  } else if (s == "OM_UNIFORM") {
    return new UniformDist(modelFile);
  } else {
    assert(false);
  }
}

void BinaryDist::Learn(const vector<doublepair> & ss)
{
  int i;
  double cnt[2] = {0, 0};

  for (i = 0; i < ss.size(); i++) {
    const doublepair & dp = ss[i];
    cnt[0] += dp.first;
    if (abs(dp.second) < g_eps) {
      cnt[1] += dp.first;
    }
  }
  _prob0 = cnt[1] / cnt[0];
}

void GaussianDist::Learn(const vector<doublepair> & ss)
{
  int i;
  double sum = 0, cnt = 0, mean, diff;

  for (i = 0; i < ss.size(); i++) {
    const doublepair & dp = ss[i];
    cnt += dp.first;
    sum += dp.first * dp.second;
  }
  mean = sum / cnt;
  sum = 0;
  for (i = 0; i < ss.size(); i++) {
    const doublepair & dp = ss[i];
    diff = dp.second - mean;
    sum += dp.first * diff * diff;
  }
  _mean = mean;
  _var = sum / cnt;
}

void ExponentialDist::Learn(const vector<doublepair> & ss)
{
  int i;
  double sum = 0, cnt = 0, mean;

  for (i = 0; i < ss.size(); i++) {
    const doublepair & dp = ss[i];
    cnt += dp.first;
    sum += dp.first * dp.second;
  }
  mean = sum / cnt;
  if (_direc == 0) { // if to the right
    _lambda = 1.0 / (mean - _shift);
  } else { // if to the left
    _lambda = 1.0 / (_shift - mean);
  }
}

MixtureDist::MixtureDist(istream & modelFile)
{
  int size, i;
  double weight, sumw = 0.0;

  _dists.clear();
  _weights.clear();
  modelFile >> _maxiter;
  modelFile >> size;
  for (i = 0; i < size; i++) {
    modelFile >> weight;
    _weights.push_back(weight);
    sumw += weight;
    _dists.push_back(ReadDistribution(modelFile));
  }
  assert(abs(sumw - 1.0) < g_eps);
}

void MixtureDist::Learn(const vector<doublepair> & ss)
{
  int iter, i, t, size = _dists.size();
  // Compute llikelihood instead of calling LogLikelihood() to speed up
  double sump, post, sumw, llikelihood, oldll = log(0.0); // -inf
  vector<double> prob(size), neww(size);
  vector<vector<doublepair> > cpntss(size, vector<doublepair>());

  for (iter = 0; iter < _maxiter; iter++) {
    for (i = 0; i < size; i++) {
      neww[i] = 0.0;
    }
    llikelihood = 0.0;
    for (t = 0; t < ss.size(); t++) {
      const doublepair & dp = ss[t];
      // To avoid numeric error to make dp.first < 0 or to make
      // sump == 0, in which case dp.first should be tiny
      if (dp.first > g_tinyeps) {
	sump = 0.0;
	for (i = 0; i < size; i++) {
	  prob[i] = _dists[i]->GetProb(dp.second) * _weights[i];
	  sump += prob[i];
	}
	llikelihood += dp.first * log(sump);
	for (i = 0; i < size; i++) {
	  // posterior of in dist[i] given dp->second
	  post = dp.first * prob[i] / sump;
	  cpntss[i].push_back(make_pair(post, dp.second));
	  neww[i] += post;
	}
      }
    }
    cout << "Log likelihood at start of iteration " << iter << ": " << llikelihood << endl;
    sumw = 0.0;
    for (i = 0; i < size; i++) {
      _dists[i]->Learn(cpntss[i]);
     sumw += neww[i];
    }
    for (i = 0; i < size; i++) {
      _weights[i] = neww[i] / sumw;
    }
    if (llikelihood - oldll < g_eps * ss.size()) {
      break;
    }
    oldll = llikelihood;
  }
  if (iter == _maxiter) {
    cout << "Warning: not converge. Too many iters: " << iter << endl;
  } else {
    cout << "Converged in " << iter << " iterations" << endl;
  }
}
