#ifndef _DISTRIBUTION_H
#define _DISTRIBUTION_H

#include <iostream>
#include <vector>
#include <cmath>
#include <cassert>
using namespace std;

typedef pair<double, double> doublepair;
// Precision for deciding whether doubles are the same
const double g_eps = 0.00001;
const double g_tinyeps = 1e-9;
const double g_pi = 3.14159;

class Distribution {
 public:
  virtual ~Distribution() {}

  virtual double GetProb(double val) const = 0;
  virtual void Learn(const vector<doublepair> & ss) = 0;
  virtual bool Assert(double val) const = 0;
  // sep: 0: one line without endl. 1: separate using endl
  virtual void Print(ostream & out, int sep = 1) const = 0;
  virtual Distribution * Duplicate() const = 0;

  inline double LogLikelihood(const vector<doublepair> & ss);
};

Distribution * ReadDistribution(istream & modelFile);

class BinaryDist : public Distribution {
 public:
  // BinaryDist(double prob0) : _prob0(prob0) {}
  inline BinaryDist(istream & modelFile);

  inline virtual double GetProb(double val) const;
  virtual void Learn(const vector<doublepair> & ss);
  inline virtual bool Assert(double val) const;
  inline virtual void Print(ostream & out, int sep = 1) const;
  inline virtual Distribution * Duplicate() const;
 private:
  double _prob0; // prob of being 0
};

class GaussianDist : public Distribution {
 public:
  // GaussianDist(double mean, double var) : _mean(mean), _var(var) {}
  inline GaussianDist(istream & modelFile);

  inline virtual double GetProb(double val) const;
  virtual void Learn(const vector<doublepair> & ss);
  inline virtual bool Assert(double val) const;
  inline virtual void Print(ostream & out, int sep = 1) const;
  inline virtual Distribution * Duplicate() const;
 private:
  double _mean, _var; // mean and variance
};

class ExponentialDist : public Distribution {
 public:
  // ExponentialDist(double shift, int direc, double lambda) :
  // _shift(shift), _direc(direc), _lambda(lambda) {}
  inline ExponentialDist(istream & modelFile);

  inline virtual double GetProb(double val) const;
  virtual void Learn(const vector<doublepair> & ss);
  inline virtual bool Assert(double val) const;
  inline virtual void Print(ostream & out, int sep = 1) const;
  inline virtual Distribution * Duplicate() const;
 private:
  double _shift, _lambda; // shift and lambda
  int _direc; // right (0) or left (1)
};

class MixtureDist : public Distribution {
 public:
  // MixtureDist() { _weights.clear(); _dists.clear(); }
  MixtureDist(istream & modelFile);
  inline virtual ~MixtureDist();

  inline virtual double GetProb(double val) const;
  virtual void Learn(const vector<doublepair> & ss);
  inline virtual bool Assert(double val) const;
  inline virtual void Print(ostream & out, int sep = 1) const;
  inline virtual Distribution * Duplicate() const;
 private:
  int _maxiter;
  // If too slow, we can fix _weights so no need to iterate in learning
  vector<double> _weights;
  vector<Distribution *> _dists;
};

class UniformDist : public Distribution {
 public:
  inline UniformDist(istream & modelFile);

  inline virtual double GetProb(double val) const;
  virtual void Learn(const vector<doublepair> & ss) {}
  inline virtual bool Assert(double val) const;
  inline virtual void Print(ostream & out, int sep = 1) const;
  inline virtual Distribution * Duplicate() const;
 private:
  double _begin, _end, _density;
  // inclusiveness at the extreme points: 0: (). 1: (]. 2: [). 3: []
  int _opcl;
};

inline double Distribution::LogLikelihood(const vector<doublepair> & ss) {
  int i;
  double ret = 0.0;
  for (i = 0; i < ss.size(); i++) {
    ret += ss[i].first * log(GetProb(ss[i].second));
  }
  return ret;
}

inline BinaryDist::BinaryDist(istream & modelFile) {
  modelFile >> _prob0;
  assert(_prob0 >= 0.0 && _prob0 <= 1.0);
}

inline double BinaryDist::GetProb(double val) const {
  if (abs(val) < g_eps) {
    return _prob0;
  } else {
    return 1 - _prob0;
  }
}

inline bool BinaryDist::Assert(double val) const {
  if (abs(val) < g_eps || abs(val - 1.0) < g_eps) {
    return true;
  } else {
    return false;
  }
}

inline void BinaryDist::Print(ostream & out, int sep) const {
  out << "OM_BINARY\t" << _prob0;
  if (sep == 1) {
    out << '\n';
  }
}

inline Distribution * BinaryDist::Duplicate() const {
  return new BinaryDist(*this);
}

inline GaussianDist::GaussianDist(istream & modelFile) {
    modelFile >> _mean >> _var; // val: variance
    assert(_var > 0.0);
}

inline double GaussianDist::GetProb(double val) const {
  double d = val - _mean;
  return exp(- d * d / _var / 2.0) / sqrt(2.0 * g_pi * _var);
}

inline bool GaussianDist::Assert(double val) const {
  return true;
}

inline void GaussianDist::Print(ostream & out, int sep) const {
  out << "OM_GAUSSIAN\t" << _mean << '\t' << _var;
  if (sep == 1) {
    out << '\n';
  }
}

inline Distribution * GaussianDist::Duplicate() const {
  return new GaussianDist(*this);
}

inline ExponentialDist::ExponentialDist(istream & modelFile) {
  modelFile >> _shift >> _direc >> _lambda;
  assert(_direc == 0 || _direc == 1);
  assert(_lambda > 0.0);
}

inline double ExponentialDist::GetProb(double val) const {
    if (_direc == 0) { // if it's to the right
      return _lambda * exp(- _lambda * (val - _shift));
    } else { // if it's to the left
      return _lambda * exp(- _lambda * (_shift - val));
    }
}

inline bool ExponentialDist::Assert(double val) const {
  if (_direc == 0) { // if it's to the right
    return (val >= _shift);
  } else { // if it's to the left
    return (val <= _shift);
  }
}

inline void ExponentialDist::Print(ostream & out, int sep) const {
  out << "OM_EXPONENTIAL\t" << _shift << '\t' << _direc << '\t' << _lambda;
  if (sep == 1) {
    out << '\n';
  }
}

inline Distribution * ExponentialDist::Duplicate() const {
  return new ExponentialDist(*this);
}

inline MixtureDist::~MixtureDist()
{
  int i;
  for (i = 0; i < _dists.size(); i++) {
    delete _dists[i];
  }
}

inline double MixtureDist::GetProb(double val) const {
  int i;
  double ret = 0.0;
  for (i = 0; i < _dists.size(); i++) {
    ret += _dists[i]->GetProb(val) * _weights[i];
  }
  return ret;
}

inline bool MixtureDist::Assert(double val) const {
  int i;
  for (i = 0; i < _dists.size(); i++) {
    if (_dists[i]->Assert(val) == false) {
      return false;
    }
  }
  return true;
}

inline void MixtureDist::Print(ostream & out, int sep) const {
  int i, size = _dists.size();
  out << "OM_MIXTURE\t" << _maxiter << '\t' << size;
  if (sep == 1) {
    out << '\n';
  }
  for (i = 0; i < size; i++) {
    if (sep == 0) {
      out << '\t';
    }
    out << _weights[i] << '\t';
    _dists[i]->Print(out, sep);
  }
}

inline Distribution * MixtureDist::Duplicate() const {
  MixtureDist * dist = new MixtureDist(*this);
  int i;
  for (i = 0; i < _dists.size(); i++) {
    dist->_dists[i] = _dists[i]->Duplicate();
  }
  return dist;
}

inline UniformDist::UniformDist(istream & modelFile) {
  modelFile >> _begin >> _end >> _opcl;
  assert(_begin < _end && _opcl >= 0 && _opcl <= 3);
  _density = 1.0 / (_end - _begin);
}

inline double UniformDist::GetProb(double val) const {
  if (val > _begin || (_opcl >> 1) == 1 && val == _begin) {
    if (val < _end || (_opcl & 1) == 1 && val == _end) {
      return _density;
    }
  }
  return 0.0;
}

inline bool UniformDist::Assert(double val) const {
  return true;
}

inline void UniformDist::Print(ostream & out, int sep) const {
  out << "OM_UNIFORM\t" << _begin << '\t' << _end << '\t' << _opcl;
  if (sep == 1) {
    out << '\n';
  }
}

inline Distribution * UniformDist::Duplicate() const {
  return new UniformDist(*this);
}

#endif // _DISTRIBUTION_H
