#include <iostream>
#include <fstream>
#include "varParaMem.h"

// --- Definitions for class ProtMot ---
ProtMot::ProtMot(int np, int nm) : numProt(np), numMot(nm)
{
  int i;

  nonbind[0] = -1;
  protMotLen = new map<int, int>[numProt];
  protLen = new int[numProt];
  motCover = new int[numProt];
  for (i = 0; i < numProt; i++) {
    protLen[i] = -1;
    motCover[i] = -1;
  }
}

void ProtMot::SetNonBind(const int prot1, const int prot2, const int mot1, int mot2)
{
  int epl, emc, before1, after1;

  nonbind[0] = prot1;
  nonbind[1] = prot2;
  nonbind[2] = mot1;
  nonbind[3] = mot2;
  switch (_leakType) {
  case 0:
  case 2:
    _numnbleak = 0;
    break;
  case 1:
    if (nonbind[3] == -1) {
      // this is # leak vars before force no binding MINUS 1
      before1 = (int)sqrt((protLen[prot1] * protLen[prot2] - motCover[prot1] * motCover[prot2]) * 10 / leakGroup);
      epl = protLen[prot1] - protMotLen[prot1][mot1];
      emc = motCover[prot1] - protMotLen[prot1][mot1];
      // this is # leak vars after force no binding MINUS 1
      after1 = (int)sqrt((epl * protLen[prot2] - emc * motCover[prot2]) * 10 / leakGroup);
      _numnbleak = before1 - after1;
    } else {
      _numnbleak = 0;
    }
    break;
  default:
    assert(0);
  }
}

int ProtMot::IsNonBind(const int prot1, const int prot2, const int mot1, const int mot2) const
{
  if (nonbind[0] == -1) {
    return 0;
  }
  if (prot1 == nonbind[0] && prot2 == nonbind[1] && mot1 == nonbind[2] && (nonbind[3] == -1 || nonbind[3] == mot2)) {
    return 1;
  }
  if (prot2 == nonbind[0] && prot1 == nonbind[1] && mot2 == nonbind[2] && (nonbind[3] == -1 || nonbind[3] == mot1)) {
    return 1;
  }
  return 0;
}

int ProtMot::NumLeak(const int prot1, const int prot2) const
{
  switch (_leakType) {
  case 0:
    return 1;
    break;
  case 1:
    return ((int)sqrt((protLen[prot1] * protLen[prot2] - motCover[prot1] * motCover[prot2]) * 10 / leakGroup) + 1);
    break;
  case 2:
    return 0;
    break;
  default:
    assert(0);
  }
}

int ProtMot::NumNBLeak(const int prot1, const int prot2) const
{
  switch (_leakType) {
  case 0:
  case 2:
    return 0;
    break;
  case 1:
    if (WithNonBind(prot1, prot2) == 1) {
      return _numnbleak;
    } else {
      return 0;
    }
    break;
  default:
    assert(0);
  }
}

void ProtMot::Print(ofstream & outFile) const
{
  int c, i;
  map<int, int>::const_iterator moti;

  outFile << "***s Motifs:" << endl;
  c = 0;
  for (i = 0; i < numProt; i++) {
    c += protMotLen[i].size();
    outFile << "Protein " << i << ":";
    for (moti = protMotLen[i].begin(); moti != protMotLen[i].end(); ++moti) {
      outFile << '\t' << moti->first << ',' << moti->second;
    }
    outFile << endl;
  }
  outFile << "***e Motifs." << endl;
  outFile << "*** Total # motifs:\t" << c << endl;
  outFile << "***s Protein lengths and motif coverage:" << endl;
  for (i = 0; i < numProt; i++) {
    outFile << i << '\t' << protLen[i] << '\t' << motCover[i] << endl;
  }
  outFile << "***e Protein lengths and motif coverage." << endl;
  outFile << "*** Forced nonbinding (prot1, prot2, mot1, mot2):\t" << nonbind[0]  << '\t' << nonbind[1] << '\t' << nonbind[2] << '\t' << nonbind[3] << endl;
  outFile << "*** (numnbleak, leakGroup size):\t" << _numnbleak << '\t' << leakGroup << endl;
}

// --- Definitions for class MotInterPot ---
void MotInterPot::Print(ostream & outFile) const
{
  map<int, double>::const_iterator mipi;

  outFile << "***s Motif pair affinity:" << endl;
  for (mipi = motInterPot.begin(); mipi != motInterPot.end(); ++mipi) {
    outFile << RevIndex(mipi->first).first << '\t' << RevIndex(mipi->first).second << '\t' << mipi->second << endl;
  }
  outFile << "***e Motif pair affinity." << endl;
  outFile << "*** Total # motif pairs:\t" << motInterPot.size() << endl;
  outFile << "*** Leak probability:\t" << leakProb << endl;
}

void MotInterPot::Cond(const MotInterPot & mip, const ObsModel & dm, const ProtInter & di)
{
  int x, y;
  double pd0, pd1, p1;
  const map<int, double> & mp = GetActPair();
  map<int, double>::const_iterator mpi;
  map<int, INTERDATA>::const_iterator idata;
  assert(numMot == mip.numMot);
  assert(mp.size() == mip.GetActPair().size());

  for (mpi = mp.begin(); mpi != mp.end(); mpi++) {
    // Have to convert to x and y first b/c no guarantee micp and di
    // will use the same indexing system
    x = RevIndex(mpi->first).first;
    y = RevIndex(mpi->first).second;
    idata = di.GetInter(x, y);
    if (idata != di.GetInterEnd()) {
      // Prob of observation given non-ddi
      pd0 = dm.GetProb(0, idata->second);
      // Prob of observation given ddi
      pd1 = dm.GetProb(1, idata->second);
    } else { // mot pair (x, y) may not have any evidence
      pd0 = 1.0;
      pd1 = 1.0;
    }
    p1 = mip.GetPotIndex(mpi->first); // Prob of ddi given nothing
    SetPotIndex(mpi->first, pd1 * p1 / (pd0 * (1 - p1) + pd1 * p1));
    // SetPotIndex(mpi->first, (pd1 * p1 + g_paramLap) / (pd0 * (1 - p1) + pd1 * p1 + g_paramLap * 2.0));
  }  
  SetLeak(mip.GetLeak());
}

// --- Definitions for class ProtInter ---
void ProtInter::AddInter(const int prot1, const int prot2, const double val, int modelid0, int modelid1)
{
  struct interdatum obsval;
  obsval.val = val;
  obsval.modelid[0] = modelid0;
  obsval.modelid[1] = modelid1;
  interPair[Index(prot1, prot2)].push_back(obsval);
}

bool ProtInter::GetInterVal(const int prot1, const int prot2, const int id[], double * val) const
{
  map<int, INTERDATA>::const_iterator ipp = interPair.find(Index(prot1, prot2));
  if (ipp == interPair.end()) {
    return false;
  }
  return GetObsVal(ipp->second, id, val);
}

void ProtInter::Print(const ProtMot & pm, ofstream & outFile) const
{
  int i, j, k;
  map<int, INTERDATA>::const_iterator ipi;

  outFile << "***s Observations (prot1, prot2, # nonmotif aa pair with all binding, # 0th leak var, (obs val, modelid1, modelid2)+):" << endl;
  for (ipi = interPair.begin(); ipi != interPair.end(); ++ipi) {
    i = RevIndex(ipi->first).first;
    j = RevIndex(ipi->first).second;
    outFile << i << '\t' << j << '\t' << pm.GetProtLen(i) * pm.GetProtLen(j) - pm.GetMotCover(i) * pm.GetMotCover(j) << '\t' << pm.NumLeak(i, j);
    const INTERDATA & obsvals = ipi->second;
    for (k = 0; k < obsvals.size(); k++) {
      outFile << '\t' << obsvals[k].val << '\t' << obsvals[k].modelid[0] << '\t' << obsvals[k].modelid[1];
    }
    outFile << endl;
  }
  outFile << "***e Observations." << endl;
  outFile << "*** Total # interacting protein pairs:\t" << interPair.size() << endl;
}

// --- Definitions for class ObsModel ---
double ObsModel::GetProb(int interval, const INTERDATA & obsvals) const
{
  int i;
  double prod = 1.0;
  for (i = 0; i < obsvals.size(); i++) {
    prod *= GetOneProb(interval, obsvals[i]);
  }
  return prod;
}

void ObsModel::M2Step(MotInterPot * micp, const MotInterPot * mip, const ProtInter * di)
{
  int modelid;
  /*
  if (_statSize.size() == 0) { // 1st time in M2Step()
    cout << "stat size:" << endl;
    for (i = 0; i < _dists.size(); i++) {
      cout << i << '\t' << _ss[i].size() << endl;
      _statSize.push_back(_ss[i].size());
    }
  } else {
    for (i = 0; i < _dists.size(); i++) {
      assert(_ss[i].size() == _statSize[i]);
    }
  }
  */
  for (modelid = 0; modelid < _dists.size(); modelid++) {
    _dists[modelid]->Learn(_ss[modelid]);
  }
  _dirty = true;
  if (micp != NULL) {
    micp->Cond(*mip, *this, *di);
  }
  ClearStat();
}

void ObsModel::Print(ostream & out) const
{
  int i;
  out << "*** # models:\t" << _dists.size() << endl;
  out << "***s Model info (stat size, ID, type, (param)+):" << endl;
  for (i = 0; i < _dists.size(); i++) {
    out << _ss[i].size() << '\t' << i << '\t';
    _dists[i]->Print(out);
  }
  out << "***e Model info." << endl;
}

double ObsModel::GetOneProb(int interval, const struct interdatum & obsval) const
{
  int modelid = obsval.modelid[interval];
  if (modelid == -1) {
    if (abs(obsval.val - interval) < g_eps) {
      return 1.0;
    } else {
      return 0.0;
    }
  } else if (modelid == -2) {
    if (obsval.val >= 0.0 && obsval.val <= 1.0) {
      return 1.0;
    } else {
      return 0.0;
    }
  }
  return _dists[modelid]->GetProb(obsval.val);
}

void ObsModel::AddOneStat(double prob1, const struct interdatum & obsval) const
{
  double prob[2] = {1 - prob1, prob1};
  int i, modelid;
  for (i = 0; i < 2; i++) {
    modelid = obsval.modelid[i];
    if (modelid >= 0) {
      _ss[modelid].push_back(make_pair(prob[i], obsval.val));
      //    } else {
      //      assert(modelid == -1 || modelid == -2);
    }
  }
}

// --- Definitions outside classes ---
int IsActPos(const INTERDATA & obsvals)
{
  int i;
  for (i = 0; i < obsvals.size(); i++) {
    if (IsOneAct(obsvals[i], 1) == 1) {
      return 1;
    }
  }
  return 0;
}

int IsActNeg(const INTERDATA & obsvals)
{
  int i;
  for (i = 0; i < obsvals.size(); i++) {
    if (IsOneAct(obsvals[i], 0) == 1) {
      return 1;
    }
  }
  return 0;
}

int IsOneAct(const struct interdatum & obsval, const int interval)
{
  if (obsval.modelid[0] == -1 && obsval.modelid[1] == -1 && abs(obsval.val - interval) < g_eps) {
    return 1;
  } else {
    return 0;
  }
}

bool GetObsVal(const INTERDATA & obsvals, const int id[], double * val)
{
  int i;
  bool ret = false;

  for (i = 0; i < obsvals.size(); i++) {
    // assume either all or none of the 2 id's match. Relax later by
    // obsvals[i].modelid[0] == id[0] && obsvals[i].modelid[1] == id[1]
    if (obsvals[i].modelid[0] == id[0]) {
      assert(obsvals[i].modelid[1] == id[1]);
      // Assume only match once. Otherwise we can return here
      assert(ret == false);
      ret = true;
      if (val != NULL) {
	(*val) = obsvals[i].val;
      }
    }
  }
  return ret;
}
