#ifndef _VARPARA_H
#define _VARPARA_H

#include "first.h"
#include "distribution.h"

class ObsModel;
class ProtInter;

// As long as we always pass by reference or create const copies,
// there is no need for copy constructor even though private member
// variables are pointers that needs to be allocated and de-allocated
class ProtMot {
 public:
  ProtMot(int np, int nm);
  // ??? can be easily extended to accept # proteins & # motifs as parameters instead
  // of using NUMPROT and NUMMOT. Same thing for the other two classes.
  inline ~ProtMot();
  inline void AddProtLen(const int prot, const int len); // assign prot to have length len
  inline void AddMotCover(const int prot, const int mc); // assign the motif coverage for prot to be mc
  inline int GetProtLen(const int prot) const; // length would be -1 before initial assignment
  inline void AddMot(const int prot, const int mot, const int len); // ??? Assume the mot is not present in the prot
  inline int IsMot(const int prot, const int mot) const; // 1: if mot is on prot. 0: otherwise
  inline const map<int, int> & GetMots(const int prot) const; // Get set of all motifs for prot
  inline int GetMotSize(const int prot) const; // # motifs on prot
  inline int GetMotCover(const int prot) const; // Get total length of all motifs on prot
  inline int GetMotLen(const int prot, const int mot) const; // Get length of mot on prot
  int GetLeakType() const { return _leakType; }
  void SetLeakType(const int leakType) { _leakType = leakType; }
  inline void SetLeakGroup(const int lg); // set leakGroup: <# a.a. pairs in leakProb>, a parameter specified by the user
  // force mot1 on prot1 to be nonbind with prot2 if mot2 == -1
  // otherwise force mot1 on prot1 nonbind w/ mot2 on prot2
  void SetNonBind(const int prot1, const int prot2, const int mot1, const int mot2 = -1);
  // remove forced nonbinding pairs so there is no forced nonbinding
  inline void UnsetNonBind();
  // return 1 if there is forced non-binding between any motif on
  // prot1 and some other motif on prot2. 0 otherwise
  inline int WithNonBind(const int prot1, const int prot2) const;
  // return 1 if mot1 on prot1 is forced to be nonbinding with mot2
  // on prot2. 0 otherwise
  int IsNonBind(const int prot1, const int prot2, const int mot1, const int mot2) const;
  // # leak variables needed between prot1 & prot2, which is the # non-motif a.a. pair divided by leakGroup
  //   which gives # 0th leak variables we need for protein pair (prot1, prot2)
  // Now take the floor of the division. May use ceiling or other rounding in the future
  int NumLeak(const int prot1, const int prot2) const;
  // # leak variables fixed to be non-binding
  int NumNBLeak(const int prot1, const int prot2) const;
  // # leak variables not fixed to be non-binding
  inline int NumActLeak(const int prot1, const int prot2) const;
  void Print(ofstream & outFile) const;
 private:
  int numProt, numMot;
  map<int, int> * protMotLen; // a map of motifs to lengths for each protein
  int * protLen; // length of proteins for computing # leak variables. It's -1 if the length is not assigned yet
  // length of all motifs on a protein. Used for computing # leak variables. It's -1 before assigned
  // It may not equal the sum of all motif lengths in protMotLen for the protein because of overlap
  int * motCover;
  // info about the forced nonbinding
  // nonbind[0]: 1st prot. nonbind[1]: 2nd prot
  // nonbind[2]: mot1 on 1st prot that's forced not to bind 2nd prot
  // nonbind[3]: mot2 on 2nd prot that's forced not to bind 2nd prot
  // nonbind[0] != nonbind[3] == -1:
  //   mot1 on 1st prot forced not to bind 2nd prot
  // nonbind[0], nonbind[3] != -1:
  //   mot1 on 1st prot forced not to bind mot2 on 2nd prot
  // nonbind[0] == -1:
  //   no forced nonbinding
  int nonbind[4];
  // _leakType:
  //   0: always 1 leak variable
  //   1: # leak vars proportional to the square root of non-motif a.a. pair
  //   2: no leak variable like in Riley et al
  // Default: 1
  // Note: variables below only meaningful for leakType 1
  // _numnbleak: precomputed # leak variables between nonbind[0] & nonbind[1] fixed to be non-binding
  // leakGroup: <# a.a. pairs in leakProb>, a parameter specified by the user
  int _leakType, _numnbleak, leakGroup;
};

class MotInterPot {
 public:
  MotInterPot(int nm) : numMot(nm) {}
  inline void AddPot(const int mot1, const int mot2, const double pot);
  // Activate the motif pair w/ index ind w/ initial potential
  inline void AddPotIndex(const int ind, const double pot);
  inline int Index(const int mot1, const int mot2) const;
  inline intpair RevIndex(const int ind) const; // first >= second
  // index: mot1 * (mot1 + 1) / 2 + mot2 for mot1 >= mot2
  inline int IsAct(const int mot1, const int mot2) const; // NOT assume mot1, mot2 ordered
  inline int IsActIndex(const int ind) const;
  inline map<int, double> & GetActPair();
  inline const map<int, double> & GetActPair() const;
  // ??? Make sure this const overloading makes sense
  // Get the map mapping from all active motif pairs to their potential
  inline void SetPot(const int mot1, const int mot2, const double val);
  // Set the active potential between mot1 and mot2 to be val
  inline void SetPotIndex(const int ind, const double val);
  inline double GetPot(const int mot1, const int mot2) const; // Assume the pair is active. NOT assume ordered
  inline double GetPotIndex(const int ind) const; // Assume the pair w/ index ind is active
  inline void DelPotIndex(const int ind); // Inactivate the motif pair with index ind
  inline void Clear(); // Clear all active motif pairs
  inline void AddUpdatePotIndex(const int ind, const double val);
  inline void SetLeak(const double leak);
  inline double GetLeak() const;
  void Print(ostream & outFile) const;
  // Constructing affinities conditioned on the domain model
  // Assume *this has the same pair structure as mip. Only need to
  // update the value of motInterPot and leakProb
  void Cond(const MotInterPot & mip, const ObsModel & dm, const ProtInter & di);
 private:
  int numMot;
  // ??? Acess cost is log of the size. Therefore if it's too slow, we can split it into
  // many maps. One for each motif like the linked list representation for a graph. It's
  // used in class ProtMot. It will be faster but use more memory. It will be much faster
  // if we want to get all motifs interacting with a certain motif.
  map<int, double> motInterPot;
  double leakProb; // between 0 and 1. For each non-motif a.a. pair
};

struct interdatum {
  double val;
  // -1 when observed is the actual value. -2 when it's uniform between 0 and 1
  // Used to index the variables in ObsModel if non-negative
  int modelid[2];
};
typedef vector<struct interdatum> INTERDATA;

// Return 1 if any of the observation is:
//   actual: (modelid[0] == modelid[1] == -1) and
//   observed value is 1
int IsActPos(const INTERDATA & obsvals);
//   if observed value is 0
int IsActNeg(const INTERDATA & obsvals);
// Return 1 if the observation is:
//   actual: (modelid[0] == modelid[1] == -1) and
//   observed value is interval
int IsOneAct(const struct interdatum & obsval, const int interval);
// Return true if there is an observation in obsvals matches id[]
// Both id[0] & id[1] has to match and in order
// Value of the matching observation is stored in val if
//   return true & val != NULL
bool GetObsVal(const INTERDATA & obsvals, const int id[], double * val = NULL);

// Since it also handles DDI data, we need to allow prot1 == prot2
class ProtInter {
 public:
  ProtInter(int np) : numProt(np) {}
  // Add an observed of val, which uses modelid0 if the actual var is 0 and uses modelid1 if the actual var is 1
  void AddInter(const int prot1, const int prot2, const double val, int modelid0, int modelid1);
  inline const map<int, INTERDATA> & GetAllInter() const; // Get set of interacting proteins
  // prot1 may == prot2 b/c also handle DDI data. NOT assume ordered
  // check against GetInterEnd() for whether the pair is observed
  inline map<int, INTERDATA>::const_iterator GetInter(const int prot1, const int prot2) const;
  inline map<int, INTERDATA>::const_iterator GetInterEnd() const;
  inline bool IsInter(const int prot1, const int prot2) const;
  // is there a PPI var over (prot1, prot2) with observation of model
  // id[]? Both id[0] & id[1] has to match and in order
  // Value of the matching observation is stored in val if
  //   return true & val != NULL
  bool GetInterVal(const int prot1, const int prot2, const int id[], double * val = NULL) const;
  inline intpair RevIndex(const int ind) const; // first >= second
  inline int Index(const int prot1, const int prot2) const;
  void Print(const ProtMot & pm, ofstream & outFile) const;
 private:
  int numProt;
  map<int, INTERDATA> interPair;
  // ??? Acess cost is log of the size. Therefore if it's too slow, we can split it into
  // many sets. One for each protein like the linked list representation for a graph.
  // It's used in class ProtMot. It will be faster but use more memory. It will be much
  // faster if we want to get all proteins interacting with a certain protein.
};

class ObsModel {
 public:
  inline ObsModel();
  inline ObsModel(const ObsModel & old);
  inline ~ObsModel();
  inline void AddDist(Distribution * dist);
  inline bool DistAssert(int modelid, double val) const;
  // prob of all noisy observations (obsvals) given actual interaction var to be interval
  double GetProb(int interval, const INTERDATA & obsvals) const;
  bool Dirty() const { return _dirty; }
  void SetDirty() const { _dirty = true; }
  void SetClean() const { _dirty = false; }
  inline void ClearStat() const;
  inline bool IsStatEmpty() const;
  // Add a pair of sufficient stat for all noisy observations (obsvals) and actual interaction has prob1 to be 1
  // Skip if the model ID is negative
  inline void AddStat(double prob1, const INTERDATA & obsvals) const;
  // Update the observation model using statistics it accumulated &
  // optionally reconstruct micp
  void M2Step(MotInterPot * micp = NULL, const MotInterPot * mip = NULL, const ProtInter * di = NULL);
  void Print(ostream & out) const;
 private:
  // prob of one noisy observation (obsval) given actual interaction var to be interval
  double GetOneProb(int interval, const struct interdatum & obsval) const;
  // Add a pair of sufficient stat for one noisy observation (obsval) and actual interaction has prob1 to be 1
  // Skip if the model ID is negative
  void AddOneStat(double prob1, const struct interdatum & obsval) const;
  vector<Distribution *> _dists;
  mutable vector<vector<doublepair> > _ss;
  // only become dirty if param changed. Doesn't matter if stat changed
  // dirty in terms of g_curjProbChange, not about micp so only set
  //   clean after LogJointProb()
  mutable bool _dirty;
};

// --- Definitions for class ProtMot ---
inline ProtMot::~ProtMot() {
  delete [] protMotLen;
  delete [] protLen;
  delete [] motCover;
}

inline void ProtMot::AddProtLen(const int prot, const int len) {
  protLen[prot] = len;
}

inline void ProtMot::AddMotCover(const int prot, const int mc) {
  motCover[prot] = mc;
}

inline int ProtMot::GetProtLen(const int prot) const {
  return protLen[prot];
}

inline void ProtMot::AddMot(const int prot, const int mot, const int len) {
  protMotLen[prot][mot] = len;
}

inline int ProtMot::IsMot(const int prot, const int mot) const {
  if (protMotLen[prot].find(mot) == protMotLen[prot].end()) {
    return 0;
  } else {
    return 1;
  }
}

inline const map<int, int> & ProtMot::GetMots(const int prot) const {
  return protMotLen[prot];
}

inline int ProtMot::GetMotSize(const int prot) const {
  return protMotLen[prot].size();
}

inline int ProtMot::GetMotCover(const int prot) const {
  return motCover[prot];
}

inline int ProtMot::GetMotLen(const int prot, const int mot) const {
  return protMotLen[prot][mot];
}

inline void ProtMot::SetLeakGroup(const int lg) {
  leakGroup = lg;
}

inline void ProtMot::UnsetNonBind() {
  nonbind[0] = -1;
}

inline int ProtMot::WithNonBind(const int prot1, const int prot2) const {
  if (nonbind[0] != -1 && (prot1 == nonbind[0] && prot2 == nonbind[1] || prot2 == nonbind[0] && prot1 == nonbind[1])) {
    return 1;
  }
  return 0;
}

inline int ProtMot::NumActLeak(const int prot1, const int prot2) const {
  return NumLeak(prot1, prot2) - NumNBLeak(prot1, prot2);
}

// --- Definitions for class MotInterPot ---
inline void MotInterPot::AddPot(const int mot1, const int mot2, const double pot) {
  motInterPot[Index(mot1, mot2)] = pot;
}

inline void MotInterPot::AddPotIndex(const int ind, const double pot) {
  motInterPot[ind] = pot;
}

inline int MotInterPot::Index(const int mot1, const int mot2) const {
  if (mot1 < mot2) {
    return (mot2 * (mot2 + 1)) / 2 + mot1;
  } else {
    return (mot1 * (mot1 + 1)) / 2 + mot2;
  }
}

inline intpair MotInterPot::RevIndex(const int ind) const {
  // The last term is to deal w/ numeric error
  int mot2, mot1 = (int)(sqrt(ind * 2 + 0.25) - 0.5 + 0.1 / numMot);
  
  mot2 = ind - (mot1 * (mot1 + 1)) / 2;
  return make_pair(mot1, mot2);
}

inline int MotInterPot::IsAct(const int mot1, const int mot2) const {
  if (motInterPot.find(Index(mot1, mot2)) == motInterPot.end()) {
    return 0;
  } else {
    return 1;
  }
}

inline int MotInterPot::IsActIndex(const int ind) const {
  if (motInterPot.find(ind) == motInterPot.end()) {
    return 0;
  } else {
    return 1;
  }
}

map<int, double> & MotInterPot::GetActPair()
{
  return motInterPot;
}

const map<int, double> & MotInterPot::GetActPair() const
{
  return motInterPot;
}

inline void MotInterPot::SetPot(const int mot1, const int mot2, const double val) {
  motInterPot[Index(mot1, mot2)] = val;
}

inline void MotInterPot::SetPotIndex(const int ind, const double val) {
  motInterPot[ind] = val;
}

inline double MotInterPot::GetPot(const int mot1, const int mot2) const {
  return (const_cast<map<int, double> &>(motInterPot))[Index(mot1, mot2)];
}

inline double MotInterPot::GetPotIndex(const int ind) const {
  return (const_cast<map<int, double> &>(motInterPot))[ind];
}

inline void MotInterPot::DelPotIndex(const int ind) {
  motInterPot.erase(ind);
}

inline void MotInterPot::Clear() {
  motInterPot.clear();
}

inline void MotInterPot::AddUpdatePotIndex(const int ind, const double val) {
  motInterPot[ind] += val;
}

inline void MotInterPot::SetLeak(const double leak) {
  leakProb = leak;
}

inline double MotInterPot::GetLeak() const {
  return leakProb;
}

// --- Definitions for class ProtInter ---
inline const map<int, INTERDATA> & ProtInter::GetAllInter() const {
  return interPair;
}

inline map<int, INTERDATA>::const_iterator ProtInter::GetInter(const int prot1, const int prot2) const {
  return interPair.find(Index(prot1, prot2));
}

inline map<int, INTERDATA>::const_iterator ProtInter::GetInterEnd() const {
  return interPair.end();
}

inline bool ProtInter::IsInter(const int prot1, const int prot2) const {
  return (interPair.find(Index(prot1, prot2)) != interPair.end());
}

inline intpair ProtInter::RevIndex(const int ind) const {
  // The last term is to deal w/ numeric error
  int prot1 = (int)(sqrt(ind * 2 + 0.25) - 0.5 + 0.1 / numProt), prot2;
  
  prot2 = ind - (prot1 * (prot1 + 1)) / 2;
  return make_pair(prot1, prot2);
}

inline int ProtInter::Index(const int prot1, const int prot2) const {
  if (prot1 < prot2) {
    return (prot2 * (prot2 + 1)) / 2 + prot1;
  } else {
    return (prot1 * (prot1 + 1)) / 2 + prot2;
  }
}

// --- Definitions for class ObsModel ---
inline ObsModel::ObsModel() : _dirty(true) {
  _dists.clear();
  ClearStat();
}

inline ObsModel::ObsModel(const ObsModel & old) : _dirty(old._dirty) {
  int i;
  _dists.clear();
  ClearStat();
  for (i = 0; i < old._dists.size(); i++) {
    _dists.push_back(old._dists[i]->Duplicate());
    assert(old._ss[i].size() == 0);
    _ss.push_back(old._ss[i]);
  }
}

inline ObsModel::~ObsModel() {
  int i;
  for (i = 0; i < _dists.size(); i++) {
    delete _dists[i];
  }
}

inline void ObsModel::AddDist(Distribution * dist) {
  _dists.push_back(dist);
  _ss.push_back(vector<doublepair>());
  _dirty = true;
}

inline bool ObsModel::DistAssert(int modelid, double val) const {
  return _dists[modelid]->Assert(val);
}

inline void ObsModel::ClearStat() const {
  int i;
  for (i = 0; i < _dists.size(); i++) {
    _ss[i].clear();
  }
  /*
  if (_statSize.size() != 0) {
    // No need for reserve() if clear() doesn't reduce the capacity
    // assert(_ss[i].capacity() >= _statSize[i]);
    for (i = 0; i < _dists.size(); i++) {
      _ss[i].reserve(_statSize[i]);
    }
  }
  */
}

inline bool ObsModel::IsStatEmpty() const {
  int i;
  for (i = 0; i < _dists.size(); i++) {
    if (_ss[i].size() != 0) {
      return false;
    }
  }
  return true;
}

inline void ObsModel::AddStat(double prob1, const INTERDATA & obsvals) const {
  int i;
  for (i = 0; i < obsvals.size(); i++) {
    AddOneStat(prob1, obsvals[i]);
  }
}

#endif // _VARPARA_H
